"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.querySummaryHash = exports.queryReceiptMapHash = exports.queryArchivedCycleByMarker = exports.queryCycleRecordsBetween = exports.queryLatestCycleRecords = exports.queryAllCycleRecords = exports.queryAllArchivedCyclesBetween = exports.queryAllArchivedCycles = exports.updateArchivedCycle = exports.updateSummaryBlob = exports.updateReceiptMap = exports.insertArchivedCycle = exports.initStorage = exports.Collection = void 0;
const Cycles_1 = require("../Data/Cycles");
const StateMetaData = require("./StateMetaData");
const tydb_1 = require("tydb");
const Logger = require("../Logger");
async function initStorage(config) {
    // Get db file location from config
    const dbFile = config.ARCHIVER_DB;
    exports.Collection = new tydb_1.Database({
        ref: dbFile,
        model: StateMetaData.ArchivedCycle,
        persistence_adapter: tydb_1.FS_Persistence_Adapter,
        autoCompaction: 10 * 30 * 1000, // database compaction every 10 cycles
    });
    await exports.Collection.createIndex({ fieldName: 'cycleMarker', unique: true });
}
exports.initStorage = initStorage;
async function insertArchivedCycle(archivedCycle) {
    Logger.mainLogger.debug('Inserting archived cycle', archivedCycle.cycleRecord.counter, archivedCycle.cycleMarker);
    try {
        await exports.Collection.insert([StateMetaData.ArchivedCycle.new(archivedCycle)]);
        Logger.mainLogger.debug('Successfully inserted archivedCycle', archivedCycle.cycleRecord.counter);
        // let updatedArchivedCycle = await Collection.find({
        //   filter: { cycleMarker: archivedCycle.cycleMarker },
        //   project: {
        //     _id: 0,
        //   },
        // })
        // let signedDataToSend = Crypto.sign({
        //   archivedCycles: updatedArchivedCycle,
        // })
        // if (updatedArchivedCycle) {
        //   if (socketServer) socketServer.emit('ARCHIVED_CYCLE', signedDataToSend)
        // }
    }
    catch (e) {
        Logger.mainLogger.error(e);
        Logger.mainLogger.error('Unable to insert archive cycle or it is already stored in to database', archivedCycle.cycleRecord.counter, archivedCycle.cycleMarker);
    }
}
exports.insertArchivedCycle = insertArchivedCycle;
async function updateReceiptMap(receiptMapResult) {
    if (!receiptMapResult)
        return;
    try {
        const parentCycle = Cycles_1.CycleChain.get(receiptMapResult.cycle);
        if (!parentCycle) {
            Logger.mainLogger.error('Unable find record with parent cycle with counter', receiptMapResult.cycle);
            return;
        }
        const existingArchivedCycle = await queryArchivedCycleByMarker(parentCycle.marker);
        if (!existingArchivedCycle) {
            Logger.mainLogger.error('Unable find existing archived cycle with marker', parentCycle.marker);
            return;
        }
        let newPartitionMaps = {};
        if (existingArchivedCycle.receipt && existingArchivedCycle.receipt.partitionMaps) {
            newPartitionMaps = { ...existingArchivedCycle.receipt.partitionMaps };
        }
        newPartitionMaps[receiptMapResult.partition] = receiptMapResult.receiptMap;
        let newPartitionTxs = {};
        if (existingArchivedCycle.receipt && existingArchivedCycle.receipt.partitionTxs) {
            newPartitionTxs = { ...existingArchivedCycle.receipt.partitionTxs };
        }
        if (receiptMapResult.txsMapEVMReceipt) {
            for (const id of Object.keys(receiptMapResult.txsMapEVMReceipt)) {
                /* eslint-disable security/detect-object-injection */
                receiptMapResult.txsMap[id] = [...receiptMapResult.txsMap[id], receiptMapResult.txsMapEVMReceipt[id]];
            }
            newPartitionTxs[receiptMapResult.partition] = receiptMapResult.txsMap;
        }
        else {
            newPartitionTxs[receiptMapResult.partition] = receiptMapResult.txsMap;
        }
        // console.log('TxsMap', receiptMapResult.txsMap)
        await exports.Collection.update({
            filter: { cycleMarker: parentCycle.marker },
            update: {
                $set: {
                    'receipt.partitionMaps': newPartitionMaps,
                    'receipt.partitionTxs': newPartitionTxs,
                },
            },
        });
        // let updatedArchivedCycle = await Collection.find({
        //   filter: { cycleMarker: parentCycle.marker },
        //   project: {
        //     _id: 0,
        //   },
        // })
        // let signedDataToSend = Crypto.sign({
        //   archivedCycles: updatedArchivedCycle,
        // })
        // if (updatedArchivedCycle) {
        //   if (socketServer) socketServer.emit('ARCHIVED_CYCLE', signedDataToSend)
        // }
    }
    catch (e) {
        Logger.mainLogger.error('Unable to update receipt maps in archived cycle');
        Logger.mainLogger.error(e);
    }
}
exports.updateReceiptMap = updateReceiptMap;
async function updateSummaryBlob(summaryBlob, cycle) {
    if (!summaryBlob)
        return;
    try {
        const parentCycle = Cycles_1.CycleChain.get(cycle);
        if (!parentCycle) {
            Logger.mainLogger.error('Unable find record with parent cycle with counter', cycle);
            return;
        }
        const existingArchivedCycle = await queryArchivedCycleByMarker(parentCycle.marker);
        if (!existingArchivedCycle) {
            Logger.mainLogger.error('Unable find existing archived cycle with marker', parentCycle.marker);
            return;
        }
        let newPartitionBlobs = {};
        if (existingArchivedCycle.summary && existingArchivedCycle.summary.partitionBlobs) {
            newPartitionBlobs = { ...existingArchivedCycle.summary.partitionBlobs };
        }
        newPartitionBlobs[summaryBlob.partition] = summaryBlob;
        await exports.Collection.update({
            filter: { cycleMarker: parentCycle.marker },
            update: { $set: { 'summary.partitionBlobs': newPartitionBlobs } },
        });
        // let updatedArchivedCycle = await Collection.find({
        //   filter: { cycleMarker: parentCycle.marker },
        //   project: {
        //     _id: 0,
        //   },
        // })
        // let signedDataToSend = Crypto.sign({
        //   archivedCycles: updatedArchivedCycle,
        // })
        // if (updatedArchivedCycle) {
        //   if (socketServer) socketServer.emit('ARCHIVED_CYCLE', signedDataToSend)
        // }
    }
    catch (e) {
        Logger.mainLogger.error('Unable to update summary blobs in archived cycle');
        Logger.mainLogger.error(e);
    }
}
exports.updateSummaryBlob = updateSummaryBlob;
async function updateArchivedCycle(marker, field, data) {
    const updateObj = {};
    updateObj[field] = data;
    await exports.Collection.update({
        filter: { cycleMarker: marker },
        update: { $set: updateObj },
    });
    // let updatedArchivedCycle = await Collection.find({
    //   filter: { cycleMarker: marker },
    //   project: {
    //     _id: 0,
    //   },
    // })
    // let signedDataToSend = Crypto.sign({
    //   archivedCycles: updatedArchivedCycle,
    // })
    // if (updatedArchivedCycle) {
    //   if (socketServer) socketServer.emit('ARCHIVED_CYCLE', signedDataToSend)
    // }
}
exports.updateArchivedCycle = updateArchivedCycle;
async function queryAllArchivedCycles(count) {
    const archivedCycles = await exports.Collection.find({
        filter: {},
        sort: {
            'cycleRecord.counter': -1,
        },
        limit: count ? count : null,
        project: {
            _id: 0,
        },
    });
    return archivedCycles;
}
exports.queryAllArchivedCycles = queryAllArchivedCycles;
async function queryAllArchivedCyclesBetween(start, end) {
    const archivedCycles = await exports.Collection.find({
        filter: {
            $and: [{ 'cycleRecord.counter': { $gte: start } }, { 'cycleRecord.counter': { $lte: end } }],
        },
        sort: {
            'cycleRecord.counter': -1,
        },
        limit: end - start + 1,
        project: {
            _id: 0,
        },
    });
    // let cycleRecords = await Collection.find({
    //   filter: {
    //     $and: [
    //       { 'cycleRecord.counter': { $gte: start } },
    //       { 'cycleRecord.counter': { $lte: end } },
    //     ],
    //   },
    //   sort: {
    //     'cycleRecord.counter': -1,
    //   },
    // })
    return archivedCycles;
}
exports.queryAllArchivedCyclesBetween = queryAllArchivedCyclesBetween;
async function queryAllCycleRecords() {
    const cycleRecords = await exports.Collection.find({
        filter: {},
        sort: {
            'cycleRecord.counter': -1,
        },
        project: {
            _id: 0,
            cycleMarker: 0,
            receipt: 0,
            data: 0,
            summary: 0,
        },
    });
    return cycleRecords.map((item) => item.cycleRecord);
}
exports.queryAllCycleRecords = queryAllCycleRecords;
async function queryLatestCycleRecords(count = 1) {
    const cycleRecords = await exports.Collection.find({
        filter: {},
        sort: {
            'cycleRecord.counter': -1,
        },
        limit: count,
        project: {
            _id: 0,
            cycleMarker: 0,
            receipt: 0,
            data: 0,
            summary: 0,
        },
    });
    return cycleRecords.map((item) => item.cycleRecord);
}
exports.queryLatestCycleRecords = queryLatestCycleRecords;
async function queryCycleRecordsBetween(start, end) {
    const cycleRecords = await exports.Collection.find({
        filter: {
            $and: [{ 'cycleRecord.counter': { $gte: start } }, { 'cycleRecord.counter': { $lte: end } }],
        },
        sort: {
            'cycleRecord.counter': -1,
        },
    });
    return cycleRecords.map((item) => item.cycleRecord);
}
exports.queryCycleRecordsBetween = queryCycleRecordsBetween;
async function queryArchivedCycleByMarker(marker) {
    const archivedCycles = await exports.Collection.find({
        filter: { cycleMarker: marker },
    });
    if (archivedCycles.length > 0)
        return archivedCycles[0];
    return undefined;
}
exports.queryArchivedCycleByMarker = queryArchivedCycleByMarker;
async function queryReceiptMapHash(counter, partition) {
    const foundArchivedCycles = await exports.Collection.find({
        filter: { 'cycleRecord.counter': counter },
    });
    if (foundArchivedCycles.length > 0) {
        if (foundArchivedCycles[0].receipt && foundArchivedCycles[0].receipt.partitionHashes) {
            return foundArchivedCycles[0].receipt.partitionHashes[partition];
        }
    }
    return undefined;
}
exports.queryReceiptMapHash = queryReceiptMapHash;
async function querySummaryHash(counter, partition) {
    const foundArchivedCycles = await exports.Collection.find({
        filter: { 'cycleRecord.counter': counter },
    });
    if (foundArchivedCycles.length > 0) {
        if (foundArchivedCycles[0].summary && foundArchivedCycles[0].summary.partitionHashes) {
            return foundArchivedCycles[0].summary.partitionHashes[partition];
        }
    }
    return undefined;
}
exports.querySummaryHash = querySummaryHash;
//# sourceMappingURL=Storage.js.map