"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fetchAccountsBySqlQuery = exports.queryAccountsBetweenCycles = exports.queryAccountCountBetweenCycles = exports.queryAccountCount = exports.queryAccounts = exports.queryLatestAccounts = exports.queryAccountByAccountId = exports.updateAccount = exports.bulkInsertAccounts = exports.insertAccount = void 0;
const db = require("./sqlite3storage");
const _1 = require(".");
const Logger = require("../Logger");
const Config_1 = require("../Config");
const serialization_1 = require("../utils/serialization");
async function insertAccount(account) {
    try {
        // Define the table columns based on schema
        const columns = ['accountId', 'data', 'timestamp', 'hash', 'cycleNumber', 'isGlobal'];
        // Construct the SQL query with placeholders
        const placeholders = `(${columns.map(() => '?').join(', ')})`;
        const sql = `INSERT OR REPLACE INTO accounts (${columns.join(', ')}) VALUES ${placeholders}`;
        // Map the `account` object to match the columns
        const values = columns.map((column) => typeof account[column] === 'object'
            ? (0, serialization_1.SerializeToJsonString)(account[column]) // Serialize objects to JSON
            : account[column]);
        // Execute the query directly (single-row insert)
        await db.run(_1.accountDatabase, sql, values);
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Successfully inserted Account', account.accountId);
        }
    }
    catch (err) {
        Logger.mainLogger.error(err);
        Logger.mainLogger.error('Unable to insert Account or it is already stored in the database', account.accountId);
    }
}
exports.insertAccount = insertAccount;
async function bulkInsertAccounts(accounts) {
    try {
        // Define the table columns based on schema
        const columns = ['accountId', 'data', 'timestamp', 'hash', 'cycleNumber', 'isGlobal'];
        // Get batch size from config or use a default value
        const BATCH_SIZE = Config_1.config.checkpoint?.batchSize || 100;
        const MAX_RETRIES = 3;
        // Process accounts in batches
        for (let i = 0; i < accounts.length;) {
            let currentBatchSize = BATCH_SIZE;
            let batch = accounts.slice(i, i + currentBatchSize);
            let retryCount = 0;
            let success = false;
            while (retryCount < MAX_RETRIES && !success) {
                try {
                    // Construct the SQL query for this batch
                    const placeholders = batch.map(() => `(${columns.map(() => '?').join(', ')})`).join(', ');
                    const sql = `INSERT OR REPLACE INTO accounts (${columns.join(', ')}) VALUES ${placeholders}`;
                    // Flatten the batch into a single list of values
                    const values = batch.flatMap((account) => columns.map((column) => typeof account[column] === 'object'
                        ? (0, serialization_1.SerializeToJsonString)(account[column]) // Serialize objects to JSON
                        : account[column]));
                    // Execute the query for this batch
                    await db.run(_1.accountDatabase, sql, values);
                    success = true;
                    if (Config_1.config.VERBOSE) {
                        Logger.mainLogger.debug(`Successfully inserted batch of ${batch.length} accounts (${i + 1}-${Math.min(i + batch.length, accounts.length)}/${accounts.length})`);
                    }
                    // Only advance the index if we had success
                    i = i + batch.length;
                }
                catch (err) {
                    retryCount++;
                    // Reduce batch size on failure
                    if (retryCount < MAX_RETRIES) {
                        // Calculate new batch size - reduce by half
                        currentBatchSize = Math.max(1, Math.floor(batch.length / 2));
                        // Recreate batch with the new size
                        batch = accounts.slice(i, i + currentBatchSize);
                        Logger.mainLogger.warn(`Retrying account batch insertion with reduced size (${currentBatchSize}). Retry ${retryCount}/${MAX_RETRIES}`);
                    }
                    else {
                        Logger.mainLogger.error(`Failed to insert batch of accounts after ${MAX_RETRIES} retries:`, err);
                        throw err; // Re-throw to be caught by outer try-catch
                    }
                }
            }
        }
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug(`Successfully inserted Accounts`, accounts.length);
        }
    }
    catch (err) {
        Logger.mainLogger.error(err);
        Logger.mainLogger.error('Unable to bulk insert Accounts', accounts.length);
    }
}
exports.bulkInsertAccounts = bulkInsertAccounts;
async function updateAccount(account) {
    try {
        const sql = `UPDATE accounts SET cycleNumber = $cycleNumber, timestamp = $timestamp, data = $data, hash = $hash WHERE accountId = $accountId `;
        await db.run(_1.accountDatabase, sql, {
            $cycleNumber: account.cycleNumber,
            $timestamp: account.timestamp,
            $data: (0, serialization_1.SerializeToJsonString)(account.data),
            $hash: account.hash,
            $accountId: account.accountId,
        });
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Successfully updated Account', account.accountId);
        }
    }
    catch (e) {
        Logger.mainLogger.error(e);
        Logger.mainLogger.error('Unable to update Account', account);
    }
}
exports.updateAccount = updateAccount;
async function queryAccountByAccountId(accountId) {
    try {
        const sql = `SELECT * FROM accounts WHERE accountId=?`;
        const dbAccount = (await db.get(_1.accountDatabase, sql, [accountId]));
        let account;
        if (dbAccount)
            account = { ...dbAccount, data: (0, serialization_1.DeSerializeFromJsonString)(dbAccount.data) };
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Account accountId', account);
        }
        return account;
    }
    catch (e) {
        Logger.mainLogger.error(e);
        return null;
    }
}
exports.queryAccountByAccountId = queryAccountByAccountId;
async function queryLatestAccounts(count) {
    if (!Number.isInteger(count)) {
        Logger.mainLogger.error('queryLatestAccounts - Invalid count value');
        return null;
    }
    try {
        const sql = `SELECT * FROM accounts ORDER BY cycleNumber DESC, timestamp DESC LIMIT ${count ? count : 100}`;
        const dbAccounts = (await db.all(_1.accountDatabase, sql));
        const accounts = [];
        if (dbAccounts.length > 0) {
            for (const dbAccount of dbAccounts) {
                accounts.push({ ...dbAccount, data: (0, serialization_1.DeSerializeFromJsonString)(dbAccount.data) });
            }
        }
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Account latest', accounts);
        }
        return accounts;
    }
    catch (e) {
        Logger.mainLogger.error(e);
        return null;
    }
}
exports.queryLatestAccounts = queryLatestAccounts;
async function queryAccounts(skip = 0, limit = 10000) {
    let dbAccounts;
    const accounts = [];
    if (!Number.isInteger(skip) || !Number.isInteger(limit)) {
        Logger.mainLogger.error('queryAccounts - Invalid skip or limit value');
        return accounts;
    }
    try {
        const sql = `SELECT * FROM accounts ORDER BY cycleNumber ASC, timestamp ASC LIMIT ${limit} OFFSET ${skip}`;
        dbAccounts = (await db.all(_1.accountDatabase, sql));
        if (dbAccounts.length > 0) {
            for (const dbAccount of dbAccounts) {
                accounts.push({ ...dbAccount, data: (0, serialization_1.DeSerializeFromJsonString)(dbAccount.data) });
            }
        }
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Account accounts', accounts ? accounts.length : accounts, 'skip', skip);
    }
    return accounts;
}
exports.queryAccounts = queryAccounts;
async function queryAccountCount() {
    let accounts;
    try {
        const sql = `SELECT COUNT(*) FROM accounts`;
        accounts = await db.get(_1.accountDatabase, sql, []);
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Account count', accounts);
    }
    if (accounts)
        accounts = accounts['COUNT(*)'];
    else
        accounts = 0;
    return accounts;
}
exports.queryAccountCount = queryAccountCount;
async function queryAccountCountBetweenCycles(startCycleNumber, endCycleNumber) {
    let accounts;
    try {
        const sql = `SELECT COUNT(*) FROM accounts WHERE cycleNumber BETWEEN ? AND ?`;
        accounts = await db.get(_1.accountDatabase, sql, [startCycleNumber, endCycleNumber]);
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Account count between cycles', accounts);
    }
    if (accounts)
        accounts = accounts['COUNT(*)'];
    else
        accounts = 0;
    return accounts;
}
exports.queryAccountCountBetweenCycles = queryAccountCountBetweenCycles;
async function queryAccountsBetweenCycles(skip = 0, limit = 10000, startCycleNumber, endCycleNumber) {
    let dbAccounts;
    const accounts = [];
    if (!Number.isInteger(skip) || !Number.isInteger(limit)) {
        Logger.mainLogger.error('queryAccountsBetweenCycles - Invalid skip or limit value');
        return accounts;
    }
    try {
        const sql = `SELECT * FROM accounts WHERE cycleNumber BETWEEN ? AND ? ORDER BY cycleNumber ASC, timestamp ASC LIMIT ${limit} OFFSET ${skip}`;
        dbAccounts = (await db.all(_1.accountDatabase, sql, [startCycleNumber, endCycleNumber]));
        if (dbAccounts.length > 0) {
            for (const dbAccount of dbAccounts) {
                accounts.push({ ...dbAccount, data: (0, serialization_1.DeSerializeFromJsonString)(dbAccount.data) });
            }
        }
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Account accounts between cycles', accounts ? accounts.length : accounts, 'skip', skip);
    }
    return accounts;
}
exports.queryAccountsBetweenCycles = queryAccountsBetweenCycles;
async function fetchAccountsBySqlQuery(sql, value) {
    const accounts = [];
    try {
        const dbAccounts = (await db.all(_1.accountDatabase, sql, value));
        if (dbAccounts.length > 0) {
            for (const dbAccount of dbAccounts) {
                accounts.push({ ...dbAccount, data: (0, serialization_1.DeSerializeFromJsonString)(dbAccount.data) });
            }
        }
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('fetchAccountsBySqlQuery', accounts ? accounts.length : accounts);
    }
    return accounts;
}
exports.fetchAccountsBySqlQuery = fetchAccountsBySqlQuery;
//# sourceMappingURL=accounts.js.map