"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.verifyTickets = exports.verifyMultiSigs = void 0;
const ethers_1 = require("ethers");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const security_1 = require("../types/security");
const Ajv = require("ajv");
const ticketSchema_1 = require("../schemas/ticketSchema");
const ajv = new Ajv({ allErrors: true });
const validateTicketSchema = ajv.compile(ticketSchema_1.ticketSchema);
function validateVerificationConfig(config) {
    if (!config.allowedTicketSigners || typeof config.allowedTicketSigners !== 'object') {
        throw new Error('Invalid allowedTicketSigners configuration');
    }
    if (typeof config.minSigRequired !== 'number' || config.minSigRequired < 1) {
        throw new Error('minSigRequired must be a positive number');
    }
    if (typeof config.requiredSecurityLevel !== 'number') {
        throw new Error('Invalid requiredSecurityLevel');
    }
}
// TODO: consider moving it to Utils.ts
function verifyMultiSigs(rawPayload, sigs, allowedPubkeys, minSigRequired, requiredSecurityLevel) {
    if (!rawPayload || !sigs || !allowedPubkeys || !Array.isArray(sigs)) {
        return { isValid: false, validCount: 0 };
    }
    if (sigs.length < minSigRequired)
        return { isValid: false, validCount: 0 };
    // no reason to allow more signatures than allowedPubkeys exist
    // this also prevent loop exhaustion
    if (sigs.length > Object.keys(allowedPubkeys).length)
        return { isValid: false, validCount: 0 };
    let validSigs = 0;
    const messageToSign = lib_types_1.Utils.safeStringify(rawPayload);
    const seen = new Set();
    for (let i = 0; i < sigs.length; i++) {
        /* eslint-disable security/detect-object-injection */
        // The sig owner has not been seen before
        // The sig owner is listed on the server
        // The sig owner has enough security clearance
        // The signature is valid
        // The signature is not a duplicate
        if (!seen.has(sigs[i].owner.toLowerCase()) &&
            allowedPubkeys[sigs[i].owner] &&
            allowedPubkeys[sigs[i].owner] >= requiredSecurityLevel &&
            ethers_1.ethers.verifyMessage(messageToSign, sigs[i].sig).toLowerCase() === sigs[i].owner.toLowerCase()) {
            validSigs++;
            seen.add(sigs[i].owner.toLowerCase());
        }
        /* eslint-enable security/detect-object-injection */
        if (validSigs >= minSigRequired)
            break;
    }
    return {
        isValid: validSigs >= minSigRequired,
        validCount: validSigs,
    };
}
exports.verifyMultiSigs = verifyMultiSigs;
function verifyTickets(tickets, config) {
    validateVerificationConfig(config);
    if (!validateTicketSchema(tickets)) {
        return {
            isValid: false,
            errors: [
                {
                    type: 'schema',
                    message: `Schema validation failed: ${ajv.errorsText(validateTicketSchema.errors)}`,
                    validSignatures: 0,
                },
            ],
        };
    }
    const errors = [];
    for (const ticket of tickets) {
        const { data, sign, type } = ticket;
        const messageObj = { data, type };
        const verificationResult = verifyMultiSigs(messageObj, sign, config.allowedTicketSigners, config.minSigRequired, config.requiredSecurityLevel);
        if (!verificationResult.isValid) {
            errors.push({
                type,
                message: `Invalid signatures for ticket type ${type}. Found ${verificationResult.validCount} valid signatures, required ${config.minSigRequired} with security level ${security_1.DevSecurityLevel[config.requiredSecurityLevel]}`,
                validSignatures: verificationResult.validCount,
            });
        }
    }
    return {
        isValid: errors.length === 0,
        errors,
    };
}
exports.verifyTickets = verifyTickets;
//# sourceMappingURL=ticketVerification.js.map